'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

require('cli-engine-config');

var _cliEngineCommand = require('cli-engine-command');

var _cliEngineCommand2 = _interopRequireDefault(_cliEngineCommand);

var _plugin = require('./plugins/plugin');

var _plugin2 = _interopRequireDefault(_plugin);

var _linked = require('./plugins/linked');

var _linked2 = _interopRequireDefault(_linked);

var _user = require('./plugins/user');

var _user2 = _interopRequireDefault(_user);

var _builtin = require('./plugins/builtin');

var _builtin2 = _interopRequireDefault(_builtin);

var _core = require('./plugins/core');

var _core2 = _interopRequireDefault(_core);

var _lodash = require('lodash.uniqby');

var _lodash2 = _interopRequireDefault(_lodash);

var _cache = require('./plugins/cache');

var _cache2 = _interopRequireDefault(_cache);

var _namespaces = require('./namespaces');

var _namespaces2 = _interopRequireDefault(_namespaces);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Plugins {
  constructor(output) {
    this.out = output;
    this.config = output.config;
    this.cache = new _cache2.default(output);

    this.builtinPlugins = new _builtin2.default(this.out);
    this.linkedPlugins = new _linked2.default(this.out);
    this.userPlugins = new _user2.default(this.out);
    this.corePlugins = new _core2.default(this.out);

    this.plugins = this.cache.fetchManagers(this.builtinPlugins, this.linkedPlugins, this.userPlugins, this.corePlugins);
  }

  get commands() {
    let commands = [];
    for (let plugin of this.plugins) {
      try {
        commands = commands.concat(plugin.commands);
      } catch (err) {
        this.out.warn(err, `error reading plugin ${plugin.name}`);
      }
    }
    return commands;
  }

  list() {
    return this.plugins;
  }

  isPluginInstalled(name) {
    return !!this.plugins.find(p => p.name === name);
  }

  findCommand(cmd) {
    for (let plugin of this.plugins) {
      let c = plugin.findCommand(cmd);
      if (c) return c;
    }
  }

  commandsForTopic(topic) {
    let commands = this.plugins.reduce((t, p) => {
      try {
        return t.concat(p.commands.filter(c => c.topic === topic).map(c => p.findCommand(c.id)));
      } catch (err) {
        this.out.warn(err, `error reading plugin ${p.name}`);
        return t;
      }
    }, []);
    return (0, _lodash2.default)(commands, 'id');
  }

  findTopic(cmd) {
    let name = cmd.split(':')[0];
    for (let plugin of this.plugins) {
      let t = plugin.findTopic(name);
      if (t) return t;
    }
    name = cmd.split(':').slice(0, 2).join(':');
    for (let plugin of this.plugins) {
      let t = plugin.findTopic(name);
      if (t) return t;
    }
  }

  findNamespace(cmd) {
    let ns = cmd.split(':')[0];
    return this.plugins.find(p => p.topics.find(t => t.topic.split(':')[0] === ns));
  }

  async install(name, tag = 'latest') {
    if (this.plugins.find(p => p.name === name && p.tag === tag)) throw new Error(`Plugin ${name} is already installed`);
    let path = await this.userPlugins.install(name, tag);
    this.clearCache(path);
  }

  async update() {
    if (this.userPlugins.list().length === 0) return;
    await this.userPlugins.update();
    this.clearCache(...this.userPlugins.list().map(p => p.path));
  }

  async uninstall(name) {
    let plugin = this.plugins.filter(p => ['user', 'link'].includes(p.type)).find(p => p.name === name);
    if (!plugin) throw new Error(`${name} is not installed`);
    switch (plugin.type) {
      case 'user':
        {
          if (!this.config.debug) this.out.action.start(`Uninstalling plugin ${name}`);
          await this.userPlugins.remove(name);
          break;
        }
      case 'link':
        {
          if (!this.config.debug) this.out.action.start(`Unlinking plugin ${name}`);
          this.linkedPlugins.remove(plugin.path);
          break;
        }
    }
    this.clearCache(plugin.path);
    this.out.action.stop();
  }

  addPackageToPJSON(name, version = '*') {
    this.userPlugins.addPackageToPJSON(name, version);
  }

  async addLinkedPlugin(p) {
    let name = this.linkedPlugins.checkLinked(p);
    if (this.plugins.find(p => p.type === 'user' && p.name === name)) {
      throw new Error(`${name} is already installed.\nUninstall with ${this.out.color.cmd(this.config.bin + ' plugins:uninstall ' + name)}`);
    }
    if (!_namespaces2.default.namespacePermitted(p, this.config)) throw _namespaces2.default.notPermittedError;

    await this.linkedPlugins.add(p);
    this.clearCache(p);
  }

  async refreshLinkedPlugins() {
    let paths = await this.linkedPlugins.refresh();
    this.clearCache(...paths);
  }

  clearCache(...paths) {
    this.cache.deletePlugin(...paths);
  }

  get topics() {
    return (0, _lodash2.default)(this.plugins.reduce((t, p) => t.concat(p.topics), []), 'topic');
  }

}
exports.default = Plugins;