'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

require('cli-engine-config');

var _http = require('cli-engine-command/lib/http');

var _http2 = _interopRequireDefault(_http);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _rwlockfile = require('rwlockfile');

var _rwlockfile2 = _interopRequireDefault(_rwlockfile);

var _fsExtra = require('fs-extra');

var _fsExtra2 = _interopRequireDefault(_fsExtra);

var _logChopper = require('log-chopper');

var _logChopper2 = _interopRequireDefault(_logChopper);

var _moment = require('moment');

var _moment2 = _interopRequireDefault(_moment);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Updater {

  constructor(output) {
    this.out = output;
    this.config = output.config;
    this.http = new _http2.default(output);
  }

  get autoupdatefile() {
    return _path2.default.join(this.config.cacheDir, 'autoupdate');
  }
  get autoupdatelogfile() {
    return _path2.default.join(this.config.cacheDir, 'autoupdate.log');
  }
  get updatelockfile() {
    return _path2.default.join(this.config.cacheDir, 'update.lock');
  }
  get binPath() {
    return process.env.CLI_BINPATH;
  }
  get updateDir() {
    return _path2.default.join(this.config.dataDir, 'tmp', 'u');
  }

  async fetchManifest(channel) {
    if (!this.config.s3.host) throw new Error('S3 host not defined');
    try {
      let url = `https://${this.config.s3.host}/${this.config.name}/channels/${channel}/${process.platform}-${process.arch}`;
      let manifest = await this.http.get(url);
      return manifest;
    } catch (err) {
      if (err.statusCode === 403) throw new Error(`HTTP 403: Invalid channel ${channel}`);
      throw err;
    }
  }

  _catch(fn) {
    try {
      fn();
    } catch (err) {
      this.out.debug(err);
    }
  }

  async update(manifest) {
    let base = this.base(manifest);

    if (!this.config.s3.host) throw new Error('S3 host not defined');

    let url = `https://${this.config.s3.host}/${this.config.name}/channels/${manifest.channel}/${base}.tar.gz`;
    let stream = await this.http.stream(url);

    _fsExtra2.default.mkdirpSync(this.updateDir);
    let dirs = this._dirs(require('tmp').dirSync({ dir: this.updateDir }).name);

    let dir = _path2.default.join(this.config.dataDir, 'client');
    let tmp = dirs.extract;

    await this.extract(stream, tmp, manifest.sha256gz);
    let extracted = _path2.default.join(dirs.extract, base);

    this._cleanup();

    let unlock = await _rwlockfile2.default.write(this.updatelockfile, { skipOwnPid: true });
    this._rename(dir, dirs.client);
    this._rename(extracted, dir);
    unlock();

    this._cleanupDirs(dirs);
  }

  extract(stream, dir, sha) {
    const zlib = require('zlib');
    const tar = require('tar-fs');
    const crypto = require('crypto');

    return new Promise((resolve, reject) => {
      let shaValidated = false;
      let extracted = false;

      let check = () => {
        if (shaValidated && extracted) {
          resolve();
        }
      };

      let fail = err => {
        this._catch(() => {
          if (_fsExtra2.default.existsSync(dir)) {
            _fsExtra2.default.removeSync(dir);
          }
        });
        reject(err);
      };

      let hasher = crypto.createHash('sha256');
      stream.on('error', fail);
      stream.on('data', d => hasher.update(d));
      stream.on('end', () => {
        let shasum = hasher.digest('hex');
        if (sha === shasum) {
          shaValidated = true;
          check();
        } else {
          reject(new Error(`SHA mismatch: expected ${shasum} to be ${sha}`));
        }
      });

      let ignore = function (_, header) {
        switch (header.type) {
          case 'directory':
          case 'file':
            return false;
          case 'symlink':
            return true;
          default:
            throw new Error(header.type);
        }
      };
      let extract = tar.extract(dir, { ignore });
      extract.on('error', fail);
      extract.on('finish', () => {
        extracted = true;
        check();
      });

      let gunzip = zlib.createGunzip();
      gunzip.on('error', fail);

      stream.pipe(gunzip).pipe(extract);
    });
  }

  _cleanup() {
    let dir = this.updateDir;
    this._catch(() => {
      if (_fsExtra2.default.existsSync(dir)) {
        _fsExtra2.default.readdirSync(dir).forEach(d => {
          let dirs = this._dirs(_path2.default.join(dir, d));

          this._remove(dirs.node);

          if ((0, _moment2.default)(_fsExtra2.default.statSync(dirs.dir).mtime).isBefore((0, _moment2.default)().subtract(24, 'hours'))) {
            this._cleanupDirs(dirs);
          } else {
            this._removeIfEmpty(dirs);
          }
        });
      }
    });
  }

  _cleanupDirs(dirs) {
    this._moveNode(dirs);

    this._remove(dirs.client);
    this._remove(dirs.extract);
    this._removeIfEmpty(dirs);
  }

  _removeIfEmpty(dirs) {
    this._catch(() => {
      if (_fsExtra2.default.readdirSync(dirs.dir).length === 0) {
        this._remove(dirs.dir);
      }
    });
  }

  _dirs(dir) {
    let client = _path2.default.join(dir, 'client');
    let extract = _path2.default.join(dir, 'extract');
    let node = _path2.default.join(dir, 'node.exe');

    return { dir, client, extract, node };
  }

  _rename(src, dst) {
    this.out.debug(`rename ${src} to ${dst}`);
    // moveSync tries to do a rename first then falls back to copy & delete
    // on windows the delete would error on node.exe so we explicitly rename
    let rename = this.config.windows ? _fsExtra2.default.renameSync : _fsExtra2.default.moveSync;
    rename(src, dst);
  }

  _remove(dir) {
    this._catch(() => {
      if (_fsExtra2.default.existsSync(dir)) {
        this.out.debug(`remove ${dir}`);
        _fsExtra2.default.removeSync(dir);
      }
    });
  }

  _moveNode(dirs) {
    this._catch(() => {
      let dirDeleteNode = _path2.default.join(dirs.client, 'bin', 'node.exe');
      if (_fsExtra2.default.existsSync(dirDeleteNode)) {
        this._rename(dirDeleteNode, dirs.node);
      }
    });
  }

  base(manifest) {
    return `${this.config.name}-v${manifest.version}-${process.platform}-${process.arch}`;
  }

  async restartCLI() {
    await _rwlockfile2.default.read(this.updatelockfile);
    _rwlockfile2.default.unreadSync(this.updatelockfile);
    const { spawnSync } = require('child_process');
    if (!this.binPath) return;
    const { status } = spawnSync(this.binPath, process.argv.slice(2), { stdio: 'inherit', shell: true });
    this.out.exit(status);
  }

  get autoupdateNeeded() {
    try {
      const stat = _fsExtra2.default.statSync(this.autoupdatefile);
      return (0, _moment2.default)(stat.mtime).isBefore((0, _moment2.default)().subtract(5, 'hours'));
    } catch (err) {
      if (err.code !== 'ENOENT') console.error(err.stack);
      return true;
    }
  }

  async autoupdate() {
    try {
      if (!this.autoupdateNeeded) return;
      _fsExtra2.default.writeFileSync(this.autoupdatefile, '');
      if (this.config.updateDisabled) await this.warnIfUpdateAvailable();
      await this.checkIfUpdating();
      let fd = _fsExtra2.default.openSync(this.autoupdatelogfile, 'a');
      if (!this.binPath) return;
      const { spawn } = require('child_process');
      spawn(this.binPath, ['update'], { stdio: [null, fd, fd] }).on('error', e => this.out.warn(e, 'autoupdate:'));
    } catch (e) {
      this.out.warn(e, 'autoupdate:');
    }
    try {
      await _logChopper2.default.chop(this.out.errlog);
    } catch (e) {
      this.out.debug(e.message);
    }
  }

  async warnIfUpdateAvailable() {
    const manifest = await this.fetchManifest(this.config.channel);
    let local = this.config.version.split('.');
    let remote = manifest.version.split('.');
    if (local[0] !== remote[0] || local[1] !== remote[1]) {
      this.out.warn(`${this.config.name}: update available from ${this.config.version} to ${manifest.version}`);
    }
  }

  async checkIfUpdating() {
    if (await _rwlockfile2.default.hasWriter(this.updatelockfile)) {
      this.out.warn(`${this.config.name}: warning: update in process`);
      await this.restartCLI();
    } else await _rwlockfile2.default.read(this.updatelockfile);
  }
}
exports.default = Updater;